import decimal
import os
import time
from time import sleep

from boto3.dynamodb.conditions import Key, Attr
from botocore.exceptions import ClientError

from properly_util_python.dynamo_helper import DynamoHelperBase, EnvironmentControlledDynamoHelper
from properly_util_python.helper_utils import dynamo_to_dict, dict_to_dynamo
from properly_util_python.properly_logging import ProperLogger


logger = ProperLogger("properly_util_python.dynamo_utils")


def save_to_dynamo(table_name, item: dict, dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper(),
                   *args, **kwargs):
    item = add_dynamo_metadata(item)
    item['id'] = str(item['id'])
    if 'namespace' in kwargs:
        item['id'] = kwargs['namespace'] + '-' + item['id']

    table = dynamodb_helper.get_dynamo_db().Table(table_name)
    response = table.get_item(
        TableName=table_name,
        Key={
            'id': item['id']
        }
    )
    if not response.get('Item') or kwargs.get('overwrite_existing'):

        if response.get('Item'):
            print('overwrite existing dynamo table: %s id: %s' % (table_name, item['id']))

        response = table.put_item(Item=dict_to_dynamo(item))

        print('Saved to dynamo table: %s id: %s' % (table_name, item['id']))
    else:
        print('Skipping, this item already exists in dynamo table: %s id: %s' % (table_name, item['id']))

    return response



def add_dynamo_metadata(data):
    data['dynamoUploadTimeStamp'] = int(round(time.time() * 1000))

    return data


def get_dynamo_items(table_name, table_filter=None,
                     dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper(), *args, **kwargs):
    if not table_filter:
        table_filter = {}
    RETRY_EXCEPTIONS = ('ProvisionedThroughputExceededException',
                        'ThrottlingException')
    table = dynamodb_helper.get_dynamo_db().Table(table_name)

    get_table_items = table.scan
    if kwargs.get('table_get_type') == 'query':
        get_table_items = table.query

    response = get_table_items(**table_filter)
    items = response['Items']

    # todo watch for memory overflow in batching all the items as once
    # https://gist.github.com/shentonfreude/8d26ca1fc93fdb801b2c
    # https://github.com/boto/boto3/issues/597#issuecomment-323982159
    retries = 0
    max_retry = 3
    while 'LastEvaluatedKey' in response and retries < max_retry:
        if 'Limit' in table_filter and len(items) >= table_filter.get('Limit'):
            break
        try:
            response = get_table_items(ExclusiveStartKey=response['LastEvaluatedKey'], **table_filter)
            items.extend(response['Items'])
            retries = 0
        except ClientError as err:
            if err.response['Error']['Code'] not in RETRY_EXCEPTIONS:
                raise
            logger.info("Throttling issues, slowing down", {"retries": retries, "numItems": len(items)})
            sleep(1.3 ** retries)
            retries += 1

    return dynamo_to_dict(items)


def get_dynamo_item(table_name, item_value, item_key='id',
                    dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper(), ):
    """
    Return the dict representation of an object from dynamo
    :param table_name:
    :param item_value:
    :param item_key:
    :return: item: dict
    """
    deal_table = dynamodb_helper.get_dynamo_db().Table(table_name)
    resp = deal_table.get_item(
        TableName=table_name,
        Key={
            item_key: item_value
        }
    )
    item = resp.get('Item')

    if not item:
        return item

    return dynamo_to_dict(item)


def query_table(table_name, filter_key=None, filter_value=None, filter_type='eq'):
    """
    Perform a query operation on the table. Can specify filter_key (col name) and its value to be filtered.
    Returns the response.
    """

    # todo replace filter if-statements with, getattr filter statement

    if filter_type == 'not_exists':
        table_filter = {
            'FilterExpression': Attr(filter_key).not_exists()
        }
    elif filter_type == 'lt':
        table_filter = {
            'FilterExpression': Attr(filter_key).lt(filter_value)
        }
    elif filter_type == 'contains':
        table_filter = {
            'FilterExpression': Attr(filter_key).contains(filter_value)
        }

    else:
        table_filter = {
            'FilterExpression': Attr(filter_key).eq(filter_value)
        }

    items = get_dynamo_items(table_name, table_filter)
    return items


def get_last_key(config_id, table_name, META_KEY_MAP,
                 dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper()):
    # remove "dev" constant when moving to environment variables
    meta_table = dynamodb_helper.get_dynamo_db().Table(table_name)

    response = meta_table.get_item(
        Key={
            META_KEY_MAP["META_ID"]: config_id,
        })
    config = response.get('Item', None)
    if config is None:
        return None

    last_evaluated_key = config.get(META_KEY_MAP["META_CONFIG_VAL"])

    return last_evaluated_key


def set_last_key(last_eval_config, table_name, val, META_KEY_MAP,
                 dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper()):
    # remove "dev" constant when moving to environment variables
    meta_table = dynamodb_helper.get_dynamo_db().Table(table_name)

    now_seconds_epoch_utc = int(time.time())

    config_object = last_eval_config.copy()
    config_object[META_KEY_MAP["META_CONFIG_VAL"]] = val
    config_object[META_KEY_MAP["TABLE_UPDATED_AT"]] = now_seconds_epoch_utc

    meta_table.put_item(Item=config_object)




# Update 2018 07 24 : TODO:
# As of 0.10.48 of properly-util-python there is a utility library for accessing this
# Craig is not updating the calling code below as it is not currently covered in tests, but,
# this is a reminder to Craig or any others that the the next meaningful change
# and test in this area should consider replacing the code below with:
#
# from properly_util_python.distributed_property import DistributedPropertyTuple
#
#  ...
# self.last_key_config = DistributedPropertyTuple(property_id="4oH5qF2QiU2M3_ZNygpVMQ",
#                                                         short_name="LAST_PROCESSED_HISTORICAL_RAW_FOR_CLEANING",
#                                                         long_description="Last evaluated key of the historical-raw table so that multiple cleaners can cooperate on different parts of the table",
#                                                         dynamo_helper=EnvironmentControlledDynamoHelper())
#
# ...
#
# last_evaluated = response.get('LastEvaluatedKey', None)
# self.last_key_config.set_property(last_evaluated)
#

#####
# Update 2018 07 27
#
# This code is duplicated around in a number of places and shouldn't be, for now just trying to remove other places.
#
# Additionally should review the encapsulating boundary (call api seems wide)
# Still not fully removed



#############
# Query for next set of items


def query_some_items(table_name, META_CONFIG, META_KEY_MAP, limit=20,
                     dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper()):
    PROPERLY_STAGE = os.environ.get('PROPERLY_STAGE', 'dev')
    meta_table_name = "{0}-{1}".format(PROPERLY_STAGE, META_KEY_MAP["META_TABLE"])

    last_evaluated = get_last_key(META_CONFIG[META_KEY_MAP["META_ID"]], meta_table_name, META_KEY_MAP)
    table = dynamodb_helper.get_dynamo_db().Table(table_name)

    # performance improved by querying for the reserved value, no retries, if throughput exceeded will pick up next time
    if last_evaluated:
        response = table.query(
            IndexName=META_CONFIG['keyName'] + '-index',
            KeyConditionExpression=Key(META_CONFIG['keyName']).eq(-1),
            Limit=limit,
            ExclusiveStartKey=last_evaluated,
        )
    else:
        response = table.query(
            IndexName=META_CONFIG['keyName'] + '-index',
            KeyConditionExpression=Key(META_CONFIG['keyName']).eq(-1),
            Limit=limit)

    last_evaluated = response.get('LastEvaluatedKey', None)
    set_last_key(META_CONFIG, meta_table_name, last_evaluated, META_KEY_MAP)

    items_to_clean = response.get('Items')
    items_to_clean = dynamo_to_dict(items_to_clean)

    return items_to_clean


def scan_some_items(table_name, META_CONFIG, META_KEY_MAP, limit=20,
                    dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper()):
    meta_table_name = "{0}-{1}".format("dev", META_KEY_MAP["META_TABLE"])

    last_evaluated = get_last_key(META_CONFIG[META_KEY_MAP["META_ID"]], meta_table_name, META_KEY_MAP)
    table = dynamodb_helper.get_dynamo_db().Table(table_name)

    # scan should almost never be used, if you are calling this know the performance behavior
    # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/bp-query-scan.html
    if last_evaluated:
        response = table.scan(
            Limit=limit,
            ExclusiveStartKey=last_evaluated,
        )
    else:
        response = table.scan(
            Limit=limit)

    last_evaluated = response.get('LastEvaluatedKey', None)
    set_last_key(META_CONFIG, meta_table_name, last_evaluated, META_KEY_MAP)

    items_to_clean = response.get('Items')
    items_to_clean = dynamo_to_dict(items_to_clean)

    return items_to_clean


def delete_dynamo_item(table_name, item_id, id_key='id', table_filter=None,
                       dynamodb_helper: DynamoHelperBase = EnvironmentControlledDynamoHelper(), *args, **kwargs):
    item = None
    if kwargs.get('return_deleted'):
        item = get_dynamo_item(table_name, item_id)

    table = dynamodb_helper.get_dynamo_db().Table(table_name)

    if table_filter:
        table.delete_item(**table_filter)
    else:
        table.delete_item(
            Key={
                id_key: item_id,
            }
        )

    print('deleting key:', item_id)
    return item



import logging
from properly_util_python.table_helper import TableHelper
from botocore.exceptions import ClientError
import traceback
from boto3.dynamodb.conditions import Key
import time
from properly_util_python.dynamo_data_utils import DynamoData

from properly_util_python.schemas import UserSchema
from properly_util_python.dynamo_helper import DynamoHelperBase

from properly_util_python.data_utils import GuidUrlSafe

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class UserAuthHelper:

    OPERATIONS_ROLE_ID = "S4jlPBglYkqccBcWody6hw"
    BASIC_ROLE_ID = "jczqf4GgkEKd1bxtTgpJOg"

    def __init__(self, dynamo_helper: DynamoHelperBase):
        self.dynamo_helper = dynamo_helper
        self.table_helper = TableHelper(self.dynamo_helper)
        self.user_table = self.table_helper.get_table(TableHelper.USER_TABLE_NAME)


    #for overview see: https://docs.aws.amazon.com/cognito/latest/developerguide/amazon-cognito-user-pools-using-tokens-with-identity-providers.html
    # for specific claims see: http://openid.net/specs/openid-connect-core-1_0.html#StandardClaims
    # one of the best summaries of relevant fields: https://medium.com/@darutk/understanding-id-token-5f83f50fa02e
    def get_user_auth_context_from_event(self, event):

        logger.info("event type is: {} ".format(type(event)))

        if event is None:
            logger.info("event is none")
            return None


        requestContext = event.get("requestContext")
        if requestContext is None:
            logger.warning("request context is none")
            return None

        authorizer = requestContext.get("authorizer")
        if authorizer is None:
            logger.warning("authorizer is none")
            return None

        claims = authorizer.get("claims")
        if claims is None:
            logger.warning("claims is none")
            return None

        email = claims.get("email")
        if ((not isinstance(email, str)) or (email == "")):
            logger.warning("email none, invalid type, or empty")
            return None

        sub = claims.get("sub")
        if ((not isinstance(sub, str)) or (sub == "")):
            logger.warning("sub none, invalid type, or empty")
            return None

        is_email_verified = claims.get("email_verified")
        if is_email_verified is None:
            logger.info("email is verified is none")
            #remove early exit, see note below
            #return None

        #default false
        is_email_verified_bool = False
        if ( isinstance(is_email_verified, bool) and (is_email_verified) ):
            #specifically boolean type and true
            logger.info("email verified found as boolean true")
            is_email_verified_bool = True
        elif (is_email_verified == "true"):
            #specifically a string containing true
            logger.info("email verified found as string 'true'")
            is_email_verified_bool = True

        if not is_email_verified_bool:
            #Previously: don't accept unconfirmed email
            #return None
            #Now: this check is weakened by allowing jwt without email verified explicitly set
            #reasoning, the connected systems only provide verified emails, but, that could change
            # facebook : https://stackoverflow.com/a/11847514/7632432
            # google : https://stackoverflow.com/a/30357942/7632432
            # amazon : can't confirm without email as currently configured:
            #           https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-email-phone-verification.html
            logger.info("strict mode would not pass login due to verified flag")
            pass


        user_auth_context = {'auth_id': sub,
                             'auth_email': email}

        return user_auth_context


    #From: https://docs.google.com/document/d/1j6hsvKSra6HkwxFRBKnfEy_wr8nCxYn0oSO8Of5SqUE/edit
    #
    #  returns a valid properly_user_id or None if there was an unauthorizied condition

    def find_or_update_verified_user(self, user_auth: dict, requested_properly_user_id: str, role: str = None):

        auth_id = user_auth.get('auth_id')
        auth_email = user_auth.get('auth_email')
        actual_properly_user_id = self.get_properly_id_from_user_auth_id(auth_id)

        if (actual_properly_user_id):
            logger.info("Existing Verified:  properly user id: {}  auth id: {}".format(actual_properly_user_id,
                                                                                       auth_id))

            if role != None:
                self.add_role_for_user(actual_properly_user_id, role)

            return actual_properly_user_id


        #was not found, so some update will be attempted
        id_to_update = requested_properly_user_id

        actual_properly_user_id = self.get_properly_id_from_user_auth_email(auth_email)
        if (actual_properly_user_id):
            logger.info("Existing found:  properly user id: {}  with email: {}".format(actual_properly_user_id,
                                                                                       auth_email))
            #use the matching id
            id_to_update = actual_properly_user_id

        actual_properly_user_id = self.create_or_update_if_unverified(id_to_update, auth_id, auth_email)
        if not actual_properly_user_id:
            logger.error("Failed to update properly user id: {}  with user auth: {}. ".format(actual_properly_user_id,
                                                                                       user_auth))
            #will occur when something is invalid about the target record to update (some other user claimed, or different email etc.)
            return None

        if role != None:
            self.add_role_for_user(actual_properly_user_id, role)

        return actual_properly_user_id


    def add_role_for_user(self, properly_user_id: str, role: str):
        if role == None:
            raise ValueError('Must provide a role to add')

        user_key = {'id': properly_user_id}

        seconds_epoch_now = int(time.time())

        update_expression = 'set updatedAt = :updatedAt, ' \
                            '#r = list_append(if_not_exists(#r, :empty_list), :role_list)'

        expression_attributes = {':updatedAt':seconds_epoch_now,
                                 ':empty_list' : [],
                                 ':role_list' : [role],
                                 ':role_str' : role}

        expression_attribute_names = {'#r' : 'roles'} # for reserved keyword

        condition_expression = 'NOT contains (#r, :role_str)'

        try:
            response = self.user_table.update_item(
                    Key=user_key,
                    UpdateExpression=update_expression,
                    ExpressionAttributeNames=expression_attribute_names,
                    ExpressionAttributeValues=expression_attributes,
                    ConditionExpression = condition_expression
                )
        except ClientError as e:
            # one possible expected 'error' - role already exists.  ignore and continue
            if e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                logger.info(f"Attempt to add existing role {role} - skipping")
                return None
            logger.error("exception traceback: {}".format(traceback.format_exc()))

            raise


    def get_properly_id_from_user_auth_id(self, auth_id: str):

        key_condition = Key('authIdCognito').eq(auth_id)


        response = self.user_table.query(
            IndexName='authIdCognito-index',
            KeyConditionExpression=key_condition,

        )
        items = response.get('Items')
        if (len(items)>1):
            logger.error("more than one user assoicated with auth id: {}".format(auth_id))
        item = None
        for item_candidate in items[:1]:
            item = item_candidate  # use the first one found

        if item is None:
            return None

        properly_user_id = item.get('id')

        return properly_user_id


    def get_properly_id_from_user_auth_email(self, auth_email: str):

        key_condition = Key('unverifiedEmail').eq(auth_email)

        response = self.user_table.query(
            IndexName='unverifiedEmail-index',
            KeyConditionExpression=key_condition,

        )
        items = response.get('Items')
        if (len(items) > 1):
            logger.error("more than one user associated with auth email: {}".format(auth_email))

        item = None
        for item_candidate in items[:1]:
            item = item_candidate  # use the first one found

        if item is None:
            return None

        properly_user_id = item.get('id')

        return properly_user_id


    def create_or_update_by_properly_id(self, requested_properly_user_id, user_auth):

        return "INVALID_ID_NOT_IMPLEMENTED"

    def get_user_from_properly_user_id(self, properly_user_id: str):

        response = self.user_table.get_item(Key={'id': properly_user_id})
        item = response.get('Item')
        return item

    def create_or_update_if_unverified(self, properly_user_id, auth_id: str, auth_email: str):
        # if we have a new cognito user but are holding on to an old properly_user_id from a previous
        # login with a different account, re-generate the requested properly_user_id so that account
        # creation in dynamo can succeed
        existing_user = self.get_user_from_properly_user_id(properly_user_id)
        if existing_user != None and existing_user.get('authIdCognito') != auth_id and existing_user.get('unverifiedEmail') != auth_email:
            logger.info(("Attempted to create account with existing properly user id {} ".format(properly_user_id)))
            properly_user_id = GuidUrlSafe.generate()
            logger.info("Generated new ID for creation: {} ".format(properly_user_id))

        seconds_epoch_now = int(time.time())

        user_key = {'id': properly_user_id}


        expression_attributes = {':updatedAt':seconds_epoch_now,
                                 ':email': auth_email,
                                 ':authIdCognito': auth_id}



        update_expression = 'set updatedAt = :updatedAt, ' \
                            'verifiedEmail = :email,'  \
                            'authIdCognito = :authIdCognito, '\
                            'unverifiedEmail = :email, '\
                            'createdAt = if_not_exists(createdAt,:updatedAt)'


        #condition represents the idea that:
        # This account is unverified or verifed to the same id (can't verify already verified accounts)
        # and if the unverfiedEmail either doesn't exist or is the same value (can't verify unverfied accounts with different email)
      #  condition_expression = '( attribute_not_exists(authCognitoId) OR (authCognitoId = :authCognitoId) ) ' \
      #                         ' AND (attribute_not_exists(unverifiedEmail) OR (unverifiedEmail = :email) )'

        condition_expression = '(  authIdCognito = :authIdCognito OR attribute_not_exists(authIdCognito)  ) ' \
                                 ' AND (unverifiedEmail = :email OR attribute_not_exists(unverifiedEmail) ) '

        try:

            response = self.user_table.update_item(
                Key=user_key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=expression_attributes,
                ConditionExpression = condition_expression,
                ReturnValues='ALL_NEW'
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'ConditionalCheckFailedException':
                return None
            logger.error("exception traceback: {}".format(traceback.format_exc()))

            raise

        return properly_user_id


    def is_user_in_role_for_offer(self, properly_user_id, role_id, offer_id):
        #todo: this is a way of implementing "operations manager for calgary can see calgary deals"
        # or "assigned appraiser can see offers they are assigned to"
        # this is not implemented yet, just using global role check until this capability is required
        raise NotImplementedError("roles for a resource is not implemented")

    #accepts None or empty user_id and role_id
    def is_user_in_role(self, properly_user_id: str, role_id: str):

        if (not isinstance(properly_user_id, str) or (properly_user_id == "" ) ):
            return False #None or empty users are in no role


        user_table = self.table_helper.get_table(TableHelper.USER_TABLE_NAME)

        response = user_table.get_item(Key={UserSchema.ID: properly_user_id},) #Consider: ConsistentRead=True

        user_dict_dynamo = response.get('Item')
        if (user_dict_dynamo is None):
            return False  # not in a role if the user doesn't exist

        if role_id == UserAuthHelper.BASIC_ROLE_ID:
            return True  # if the user exists they are automatically in the basic role

        user_dict = DynamoData.to_dict(user_dict_dynamo)

        roles_array = user_dict.get(UserSchema.ROLES)

        if roles_array is None:
            return False

        has_role_flag = role_id in roles_array

        return has_role_flag






import base64
import json
import os
from typing import Optional

import boto3
from botocore.exceptions import ClientError


def get_secret(
        secret_name: str,
        stage: str = None,
) -> dict:
    """
    Gets a secret from the AWS Secrets Manager
    It will look for the secret with the name in the following pattern: `{stage}-{secret name}`

    :param secret_name: The name of the secret in AWS Secrets Manager.
    :param stage: (Optional) The stage name of the system environment.
                  Defaults to the environment variable "PROPERLY_STAGE".
    :return: The dictionary of secrets.
    """

    if not stage:
        stage = os.environ["PROPERLY_STAGE"]

    secret_name = f"{stage}-{secret_name}"
    secret = get_from_secrets_manager(secret_name)

    return json.loads(secret)


def get_from_secrets_manager(secret_name: str) -> Optional[str]:
    """
    Gets the secret from AWS Secrets Manager.
    This code is mostly supplied by AWS.

    :param secret_name: The name of the secret.
    :return: The string representation of the secret value.
    """
    region_name = "us-east-1"

    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(
        service_name="secretsmanager",
        region_name=region_name
    )

    # In this sample we only handle the specific exceptions for the "GetSecretValue" API.
    # See https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
    # We rethrow the exception by default.

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "DecryptionFailureException":
            # Secrets Manager can"t decrypt the protected secret text using the provided KMS key.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response["Error"]["Code"] == "InternalServiceErrorException":
            # An error occurred on the server side.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response["Error"]["Code"] == "InvalidParameterException":
            # You provided an invalid value for a parameter.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response["Error"]["Code"] == "InvalidRequestException":
            # You provided a parameter value that is not valid for the current state of the resource.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response["Error"]["Code"] == "ResourceNotFoundException":
            # We can't find the resource that you asked for.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
    else:
        # Decrypts secret using the associated KMS CMK.
        # Depending on whether the secret is a string or binary, one of these fields will be populated.
        if "SecretString" in get_secret_value_response:
            secret = get_secret_value_response["SecretString"]
            return secret

        secret = base64.b64decode(get_secret_value_response["SecretBinary"])
        return secret



import decimal
import json
import uuid
import base64
import datetime
import importlib


class GuidUrlSafe:
    @classmethod
    def generate(cls):
        id_uuid = uuid.uuid4()
        id_string = base64.urlsafe_b64encode(id_uuid.bytes).decode("utf-8")
        start_of_padding = id_string.find('=')
        if (start_of_padding >= 0):
            id_string = id_string[:start_of_padding]

        return id_string


class ExtendedJsonEncoder(json.JSONEncoder):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This code permits the class to handle pandas timestamp when the module is present
        # and skip it when pandas is not present.
        #
        # Consider: This seems like a ton of work for a constructor
        #
        self.pandas_dynamic = None

        try:
            self.pandas_dynamic = importlib.import_module('pandas')
        except ImportError as err:
            # pandas not found, leave as null so it can be skipped below
            pass

        self.np_dynamic = None
        try:
            self.np_dynamic = importlib.import_module('numpy')
        except ImportError as err:
            # numpy not found, leave as null so it can be skipped below
            pass

    def default(self, o):
        if isinstance(o, decimal.Decimal):
            if abs(o % 1) > 0:  # without abs() negative float values were being converted to int
                return float(o)
            else:
                return int(o)
        elif isinstance(o, set):
            return list(o)
        elif ((self.pandas_dynamic is not None) and ( isinstance(o, self.pandas_dynamic.Timestamp)) ):
            posix_float_timestamp = o.timestamp()
            timestamp_as_millis = int(1000.0 * posix_float_timestamp)
            return int(timestamp_as_millis)
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif (self.np_dynamic is not None):
            np =self.np_dynamic
            if isinstance(o, (np.int_, np.intc, np.intp, np.int8,
                              np.int16, np.int32, np.int64, np.uint8,
                              np.uint16, np.uint32, np.uint64)):
                return int(o)
            elif isinstance(o, (np.float_, np.float16, np.float32,
                                np.float64)):
                return float(o)
            elif isinstance(o, (np.ndarray,)):  #### This is the fix
                return o.tolist()

        return super(ExtendedJsonEncoder, self).default(o)



import boto3
import os
import json

from enum import Enum

from properly_util_python.properly_logging import ProperLogger


class ProperlyEvents(str, Enum):
    USER_CREATED = 'userCreated'
    STAGE_CHANGED_OFFER = 'stageChangedOffer'
    USER_ENTERED_ALL_OFFER_INFO = 'userEnteredAllOfferInfo'
    SHOWING_UPDATED_OR_CREATED = 'userShowingUpdatedOrCreated'
    HUBSPOT_DEAL_CHANGED = 'hubspotDealChanged'
    DEAL_CHANGED_STAGE = 'dealChangedStage'
    ACCEPTED_OFFER = 'userAcceptedOffer'
    SEARCH_RESULTS_CHANGED = 'searchResultsChanged'
    PPR_FEEDBACK = 'pprFeedback'
    FAVOURITE_ADDED = 'favouriteAdded'
    FAVOURITE_REMOVED = 'favouriteRemoved'
    USER_DEAL_STAGES_CHANGED = 'userDealStagesChanged'
    BEHAVIOUR_TRACK_EVENT = 'behaviourTrackEvent'


class EventBusBase:
    def send_event(self, properly_user_id: str, event_name: ProperlyEvents, data_block_name: str, data_block: dict):
        raise NotImplementedError


class PropertyEvents(str, Enum):
    PROPERTY_VERIFIED_INFO_UPDATED = 'propertyVerifiedInfoUpdated'
    PPR_ESTIMATE_UPDATE = 'pprEstimateUpdated'


class HistoricalPropertyEvents(str, Enum):
    HISTORICAL_PROPERTY_UPDATED = "historicalPropertyUpdated"


class NamedEventBus(EventBusBase):
    def __init__(self, stream_name, kinesis=None, logger=None):
        stage = os.environ.get("PROPERLY_STAGE", "staging")
        self.kinesis = kinesis or boto3.client("kinesis")
        self.stream_name = f"{stage}-{stream_name}"
        self.logger = logger if logger else ProperLogger("NamedEventBus")

    def send_event(self, partition_key: str, event_name: str, data_block_name: str, data_block: dict):
        event_to_send = {
            "eventName": event_name,
            data_block_name: data_block
        }
        self.logger.debug("Sending event", {"event": event_to_send})

        event_as_json = json.dumps(event_to_send)
        event_as_bytes = event_as_json.encode('utf-8')

        # Kinesis data payloads are in bytes in base64 encoded
        self.logger.debug("Event as bytes", {"event_as_bytes": event_as_bytes})
        self.kinesis.put_record(
            StreamName=self.stream_name,
            Data=event_as_bytes,
            PartitionKey=partition_key,
        )


class PropertyEventBus(NamedEventBus):
    """
    This the wrapped up implementation of the event bus for Property records.
    """
    def send_event(self, property_id: str, event_name: PropertyEvents, data_block_name: str, data_block: dict):
        """
        The send event for the Property related buses.
        :param property_id: The Property ID of the record. This is used as the bus partition key.
        :param event_name: The name of the bus event.
        :param data_block_name: The attribute name of the data block.
        :param data_block: The block of data put on the bus.
        """
        super().send_event(
            partition_key=property_id,
            event_name=event_name,
            data_block_name=data_block_name,
            data_block=data_block)


class EventBus(NamedEventBus):
    def __init__(self, kinesis=None, logger=None):
        """
        The default generic event bus handler for "events-002" Kinesis stream.
        """
        super().__init__(stream_name="events-002", kinesis=kinesis, logger=logger)

    def send_event(self, properly_user_id: str, event_name: ProperlyEvents, data_block_name: str, data_block: dict):
        """
        The send event for the default user related buses.
        :param properly_user_id: The user ID of the record.  This is used as the partition key.
        :param event_name: The name of the bus event.
        :param data_block_name: The attribute name of the data block.
        :param data_block: The block of data put on the bus.
        """
        super().send_event(
            partition_key=properly_user_id,
            event_name=event_name,
            data_block_name=data_block_name,
            data_block=data_block)



from dynamodb_json import json_util as dynamodb_util


class DynamoStreamEventHandler:
    """
    This handler reads the stream event and figures out what has changed.
    """
    @staticmethod
    def is_dynamo_stream_event(event):
        """
        Determines if the incoming event is a DynamoDB stream event.
        :param event: The incoming event
        :return: `True` if the event has records. `False` otherwise.
        """
        return "Records" in event

    def __init__(self, event, attribute_path=None):
        """
        The constructor to handle incoming and interpret DynamoDB Stream events.
        :param event: The incoming DyanmoDB stream event.
        :param attribute_path: An array with the path to the content the caller is watching.
                               Defaults to root if not provided.
        """
        self.records = event.get("Records", [])
        self.attribute_path = attribute_path or []

    def get_changed_record_information(self):
        """
        Returns a list of details about what has changed with the incoming records.
        :return: A dictionary with the modified and deleted records in the following shape:
        ```
        {
            "inserts": [ # inserted records ],
            "modifications": [ # delta records ],
            "deletions": [ # deletion records ],
        }
        ```

        For "inserts" and "modifications", it is a list of deltas.  Each delta is structured in the following way:
        ```
        {
            "value": { # The inserted/modified record in normalized JSON },
            "modified_columns": [ # The list of attributes that were changed on this record ]
            "source_table_name": # The name of the table this value came from
        }
        ```

        * The `value` is a normalized JSON and NOT in the DynamoDB JSON format.
        * The `modified_columns` includes columns that were removed, added, or whose value was modified.
        * The `source_table_name`, for Properly, includes the environment of the table.

        For "deletions", it is a list of deletions.  Each deletion is structured in the following way:
        ```
        {
            "value": { # The original record in normalized JSON },
            "modified_columns": [ # The list of attributes that were changed on this record. That is, all the columns. ]
            "source_table_name": # The name of the table this value came from
        }
        ```

        * The `value` is a normalized JSON and NOT in the DynamoDB JSON format.
        * The `modified_columns` includes all the columns that were deleted.
        * The `source_table_name`, for Properly, excludes the environment of the table.

        """
        record_inserts = []
        record_deltas = {}
        record_deletions = []
        for record in self.records:
            event_name = record.get("eventName")
            if event_name == "REMOVE":
                # Record was removed. Create a deletion record.
                record_deletion = self._get_record_deletion(record)
                record_deletions.append(record_deletion)
            elif event_name == "MODIFY":
                record_key = self._get_record_key(record)
                record_delta = self._get_record_delta(record)
                record_deltas[record_key] = self._merge_deltas(record_deltas.get(record_key), record_delta)
            else:
                record_insert = self._get_record_delta(record)
                record_inserts.append(record_insert)

        return {
            "inserts": record_inserts,
            "modifications": list(record_deltas.values()),
            "deletions": record_deletions,
        }

    def _get_record_key(self, record):
        """
        Given a DynamoDB stream record
        (https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_streams_StreamRecord.html),
        returns a string representing the affected record's primary key.
        Only supports records with partition keys (no sort keys).
        {
            ...
            'eventName': 'MODIFY',
            'dynamodb': {
                'Keys': {
                    'id': {
                        'S': 'SOMEID'
                    }
                },
            }
        }
        """
        diff_record = record.get("dynamodb")
        record_keys = list(diff_record.get("Keys").values())

        if len(record_keys) > 1:
            raise Exception('cannot handle records with composite primary keys')

        normalized_key_value = dynamodb_util.loads(record_keys[0])

        return normalized_key_value

    def _merge_deltas(self, existing_delta, new_delta):
        """
        Merge two delta objects representing subsequent changes to the same record. Returns a new delta object
        representing the combined changes.
        :param existing_delta: The current delta object, if present. May be None.
        :param new_delta: a new delta to "merge in" to the existing delta.
        :return: a delta object representing the combination of all changes reflected in existing_delta and new_delta.
        """
        combined_modified_columns = new_delta.get("modified_columns")
        if existing_delta is not None:
            existing_delta_modified_columns = existing_delta.get("modified_columns") or []
            combined_modified_columns = sorted(list(set(combined_modified_columns + existing_delta_modified_columns)))

        return {
            "value": new_delta.get("value"),
            "modified_columns": combined_modified_columns,
            "source_table_name": new_delta.get("source_table_name"),
        }

    def _get_record_delta(self, record):
        """
        Calculates the delta for a single DynamoDB record
        :param record: A single DynamoDB record
        :return: A dictionary representing the delta.  The delta is structured in the following way:
        ```
        {
            "value": { # The updated record in normalized JSON },
            "modified_columns": [ # The list of attributes that were changed on this record ]
            "source_table_name": # The name of the table this value came from
        }
        ```

        * The `value` is a normalized JSON and NOT in the DynamoDB JSON format.
        * The `modified_columns` includes columns that were removed, added, or whose value was modified.
        * The `source_table_name`, for Properly, excludes the environment of the table.
        """
        diff_record = record.get("dynamodb")
        event_name = record.get("eventName")
        if event_name == "MODIFY":
            # Make this output consistent for unit testing purposes
            modified_columns = sorted(list(self._get_modified_columns_set(record)))
        else:
            # This is for the INSERT case. All the columns are new!
            # Traverse down the path and scoop up all the keys at the end of travel.
            new_record = dynamodb_util.loads(diff_record.get("NewImage"))
            for attribute_name in self.attribute_path:
                # Traverse down the full attribute path
                new_record = new_record.get(attribute_name, {})
            modified_columns = list(new_record.keys())

        return {
            "value": self._get_new_value(record),
            "modified_columns": modified_columns,
            "source_table_name": self._get_table_name(record),
        }

    def _get_record_deletion(self, record):
        """
        Calculates the deletion for a single DynamoDB record
        :param record: A single DynamoDB record
        :return: A dictionary representing the deletion.  The deletion is structured in the following way:
        ```
        {
            "value": { # The deleted record in normalized JSON },
            "modified_columns": [ # The list of attributes that were changed on this record. That is, all the columns. ]
            "source_table_name": # The name of the table this value came from
        }
        ```

        * The `value` is a normalized JSON and NOT in the DynamoDB JSON format.
        * The `modified_columns` includes all the columns that were deleted.
        * The `source_table_name`, for Properly, excludes the environment of the table.
        """
        diff_record = record.get("dynamodb")
        old_record = dynamodb_util.loads(diff_record.get("OldImage"))
        for attribute_name in self.attribute_path:
            # Traverse down the full attribute path
            old_record = old_record.get(attribute_name, {})
        modified_columns = list(old_record.keys())

        return {
            "value": self._get_old_value(record),
            "modified_columns": modified_columns,
            "source_table_name": self._get_table_name(record),
        }

    def _get_old_value(self, record):
        """
        Returns the normalized JSON for the *old* value. This is *not* in DynamoDB JSON format.
        :param record: A single DynamoDB record
        :return: The new record in normalized JSON.
        """
        return self._get_image_by_key(record, "OldImage")

    def _get_new_value(self, record):
        """
        Returns the normalized JSON for the *new* value. This is *not* in DynamoDB JSON format.
        :param record: A single DynamoDB record
        :return: The new record in normalized JSON.
        """
        return self._get_image_by_key(record, "NewImage")

    @staticmethod
    def _get_image_by_key(record, image_key):
        """
        Returns the record based on the image key.
        :param record: A single DynamoDB record
        :param image_key: The new record in normalized JSON.
        :return:
        """
        diff_record = record.get("dynamodb")
        new_record = diff_record.get(image_key)
        json_record = dynamodb_util.loads(new_record)
        return json_record

    def _get_modified_columns_set(self, record):
        """
        Calculates which columns have been removed, added, or modified.
        :param record: A single DynamoDB record
        :return: A list of column names that have changed.
        """
        diff_record = record.get("dynamodb")

        new_record = dynamodb_util.loads(diff_record.get("NewImage"))
        old_record = dynamodb_util.loads(diff_record.get("OldImage", {}))
        for attribute_name in self.attribute_path:
            # Traverse down the full attribute path
            new_record = new_record.get(attribute_name, {})
            old_record = old_record.get(attribute_name, {})

        changed_key_set = set([])
        for key in new_record:
            if key not in old_record:
                # This column is in the new record, but not the old record
                # Therefore, this column was added to the record
                changed_key_set.add(key)

            if new_record.get(key) != old_record.get(key):
                # The column exists in both records, but changed
                # Therefore, this column was modified
                changed_key_set.add(key)

        for key in old_record:
            # Go through the old record columns
            # If a column exists in the old record, but not in the new record
            # Therefore, the column was removed from the record
            # Note: We do not need to capture changed values because it was captured above
            if key not in new_record:
                changed_key_set.add(key)

        return changed_key_set

    @staticmethod
    def _get_table_name(record):
        """
        The source ARN is in the format of:
        'arn:aws:dynamodb:us-east-1:529943178829:table/staging-listing-clean-001/stream/2018-10-18T19:46:17.718'

        We are doing the following:
        * looking for the first backslash '/' and second backslash '/'
        * then parsing out the environment out of the string

        :param record: The DynamoDB stream record
        :return: The environment agnostic table name
        """

        source_arn = record.get("eventSourceARN")

        first_index = source_arn.index("/", 0)
        second_index = source_arn.index("/", first_index+1)
        full_table_name = source_arn[first_index+1:second_index]

        first_dash_index = full_table_name.index("-", 0)
        simplified_table_name = full_table_name[first_dash_index+1:]
        return simplified_table_name



import functools

from flask import request
from flask.wrappers import Response

from properly_util_python.dynamo_helper import EnvironmentControlledDynamoHelper
from properly_util_python.user_auth import UserAuthHelper


def authorized(role):
    def outer(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):

            user_auth_helper = UserAuthHelper(EnvironmentControlledDynamoHelper())

            unauth_response = Response(
                response="{}", status=403, mimetype="application/json"
            )

            event_object = request.environ.get("event")
            user_auth = user_auth_helper.get_user_auth_context_from_event(event_object)
            if user_auth is None:
                return unauth_response

            properly_user_id_caller = user_auth_helper.get_properly_id_from_user_auth_id(
                user_auth["auth_id"]
            )
            if properly_user_id_caller is None:
                return unauth_response

            is_in_role_flag = user_auth_helper.is_user_in_role(
                properly_user_id_caller, role
            )

            if not is_in_role_flag:
                return unauth_response

            return func(*args, **kwargs)

        return inner

    return outer


def authorized_with_id(role):
    def outer(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):

            user_auth_helper = UserAuthHelper(EnvironmentControlledDynamoHelper())

            unauth_response = Response(
                response="{}", status=403, mimetype="application/json"
            )

            event_object = request.environ.get("event")
            user_auth = user_auth_helper.get_user_auth_context_from_event(event_object)
            if user_auth is None:
                return unauth_response

            properly_user_id_caller = user_auth_helper.get_properly_id_from_user_auth_id(
                user_auth["auth_id"]
            )
            if properly_user_id_caller is None:
                return unauth_response

            is_in_role_flag = user_auth_helper.is_user_in_role(
                properly_user_id_caller, role
            )

            if not is_in_role_flag:
                return unauth_response

            return func(*args, user_id=properly_user_id_caller, **kwargs)

        return inner

    return outer




from properly_util_python import helper_utils


class DynamoData:

    @classmethod
    def to_dict(cls, dynamo_data):

       return helper_utils.dynamo_to_dict(dynamo_data)

    @classmethod
    def from_dict(cls, generic_dict):

        return helper_utils.dict_to_dynamo(generic_dict)



import warnings

from boto3.dynamodb.conditions import Key
from properly_util_python.dynamo_utils import get_dynamo_item, get_dynamo_items
from properly_util_python.dynamo_helper import DynamoHelperBase
from properly_util_python.table_helper import TableHelper


class OffersFromProperty:

    def __init__(self, dynamo_helper: DynamoHelperBase):
        self.dynamo_helper = dynamo_helper
        self.table_helper = TableHelper(dynamo_helper)

        # Initialize lazy loading offers
        self.property_id = None
        self.property_item = None

    def get_offers_with_property_id(self, property_id):
        """
        Returns all relevant offers for the property ID
        :param property_id: The Property ID
        :return: List of Offers
        """
        property_item = self.get_property(property_id)
        offer_items = self._get_dynamo_offer_items_with_property_id(TableHelper.OFFER_TABLE_NAME,property_id)

        return offer_items

    def _get_dynamo_offer_items_with_property_id(self,offer_table,property_id):

        table_name_with_environment = self.table_helper.get_table_name(offer_table)
        table_filter = {
            'IndexName': 'propertyId-index',
            'KeyConditionExpression': Key('propertyId').eq(property_id)
        }

        return get_dynamo_items(table_name_with_environment, table_filter, table_get_type='query')

    def _get_dynamo_property_item_with_property_id(self,property_table,property_id):

        table_name_with_environment = self.table_helper.get_table_name(property_table)

        return get_dynamo_item(table_name_with_environment, property_id)

    def get_property(self, property_id):
        """
        Gets the Property by ID.
        This will get lazy loaded once per ID.
        If subsequent other calls uses the same property ID, it will use the lazy loaded instance.
        This saves multiple calls to the database and allows easy access to the offer from multiple places.
        :param property_id: The ID of the property.
        :return: The property dictionary
        """
        if not self.property_item or self.property_id != property_id:
            self.property_id = property_id
            self.property_item = self._get_dynamo_property_item_with_property_id(TableHelper.PROPERTY_TABLE_NAME,property_id)

        return self.property_item




import logging
import os

from googlemaps import Client

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class GoogleGeocodeClient:
    """
    Client that calls the Google geocode API to get the location details of a given address
    """

    def __init__(self,
                 google_maps_key: str = None):
        self.google_maps_key = google_maps_key or os.environ.get('GOOGLE_MAPS_KEY')

    def is_place_a_street_address(self, geocode_result: dict):
        """
        Verifies the input Google geocode result is actually address.
        By inspection, the types that are home addresses are:
        - `street_address`
        - `premise`
        - `subpremise` (This includes apartments and condos)
        :param geocode_result: The individual Google geocoding response entry.
        :return: <b>True</b> If the geocoding result is a home address. <b>False</b> otherwise.
        """
        result_types = geocode_result.get("types", None)
        if result_types is None:
            return False

        google_address_types = [
            "street_address",
            "premise",
            "subpremise",
        ]

        # Are any of the keys in address_types inside the geocode result types.
        is_address = any(key in result_types for key in google_address_types)

        return is_address

    def geocode(self, address: str) -> list:
        logger.info("Google geocoding address: {}".format(address))

        google_maps_client = Client(key=self.google_maps_key)

        geocode_results = google_maps_client.geocode(address)
        logger.info("Google geocoding response length: {}".format(len(geocode_results)))

        filtered_results = list(filter(self.is_place_a_street_address, geocode_results))
        logger.info("Google geocoding filtered response length: {}".format(len(filtered_results)))

        return filtered_results




import traceback
import boto3

import os

sqs = boto3.client('sqs', )




def add_to_queue(queue_name, message_attributes: dict, message_body=''):
    response = None
    try:
        if not queue_exists(queue_name):
            print('%s queue does not exist. Making a new queue' % queue_name)
            response = sqs.create_queue(
                QueueName=queue_name,
                Attributes={
                    'MessageRetentionPeriod': '345600'  # 4 days
                }
            )
        else:
            response = sqs.get_queue_url(QueueName=queue_name)

        queue_url = response['QueueUrl']

        sqs_message_attributes = {}

        for k, v in message_attributes.items():
            sqs_message_attributes[k] = {
                'DataType': 'String',
                'StringValue': str(v),
            }

        response = sqs.send_message(
            QueueUrl=queue_url,
            DelaySeconds=10,
            MessageAttributes=sqs_message_attributes,
            MessageBody=message_body
        )
    except Exception as e:
        print('ClientError e:', e)
        print('traceback', traceback.format_exc())
        return {'error': str(e)}

    return response


def get_from_queue(queue_name, messages_num=10):
    try:
        response = sqs.get_queue_url(QueueName=queue_name)
        queue_url = response['QueueUrl']

        response = sqs.receive_message(
            QueueUrl=queue_url,
            AttributeNames=[
                'All'
            ],
            MaxNumberOfMessages=messages_num,
            MessageAttributeNames=[
                'All'
            ],
            # Do NOT set visibility and wait timeout to zero
            # visibility timeout of zero makes the message immediately available for other consumers
            # doing this increases duplicate processing
        )

        return response

    except Exception as e:
        # print('ClientError e:', e)
        print('ClientError e:', e)
        print('traceback', traceback.format_exc())
        return None


def delete_from_queue(queue_name, receipt_handle):
    response = sqs.get_queue_url(QueueName=queue_name)
    sqs.delete_message(
        QueueUrl=response['QueueUrl'],
        ReceiptHandle=receipt_handle
    )


def queue_exists(queue_name):
    try:
        response = sqs.get_queue_url(QueueName=queue_name)

        return response['QueueUrl'] is not None
    except Exception as e:
        print('ClientError e:', e)
        print('traceback', traceback.format_exc())
        return False

def queue_full_name(queue_name):
    stage = os.environ.get('PROPERLY_STAGE', 'dev')
    queue_name_to_return = "{}-{}".format(stage, queue_name)
    return queue_name_to_return






from properly_util_python.dynamo_helper import DynamoHelperBase
from properly_util_python.table_helper import TableHelper
from properly_util_python.dynamo_data_utils import DynamoData


class OfferUtils:

    def __init__(self, dynamo_helper: DynamoHelperBase):

        self.dynamo_helper = dynamo_helper
        self.table_helper = TableHelper(self.dynamo_helper)

    def get_property_and_user_from_offer(self, offer_dict):
        property_table = self.table_helper.get_table(TableHelper.PROPERTY_TABLE_NAME)
        user_table = self.table_helper.get_table(TableHelper.USER_TABLE_NAME)

        property_id = offer_dict.get('propertyId')
        user_id = offer_dict.get('ownerId')
        if property_id is None or user_id is None:
            return None, None

        response = property_table.get_item(Key={'id': property_id})
        item = response.get('Item')
        property_dict = DynamoData.to_dict(item)

        response = user_table.get_item(Key={'id': user_id})
        item = response.get('Item')
        user_dict = DynamoData.to_dict(item)

        return property_dict, user_dict

    def build_offer_with_property_and_user(self, offer_dict: dict, property_dict: dict, user_dict: dict):

        # Consider adding filtering similar to:
        # https://github.com/GoProperly/hig-offer-api/blob/78e21aa0b962ee859a4e9e6a941f79612899f56b/offer/src/offer_handler.py#L237

        if offer_dict is None or property_dict is None or user_dict is None:
            msg = 'Non null required offer: {} property: {} user: {} '.format(offer_dict, property_dict, user_dict)
            raise RuntimeError(msg)

        offer_dict['property'] = property_dict
        offer_dict['user'] = user_dict

        return offer_dict





import logging
import os
from typing import Any, Dict, Optional, Union

import orjson


STANDARD_OFFER_ID = "offer_id"
STANDARD_USER_ID = "user_id"
STANDARD_PROPERTY_ID = "property_id"
STANDARD_LISTING_ID = "listing_id"
STANDARD_GOOGLE_PLACE_ID = "google_place_id"

STANDARD_FORMAT_MAPPER = {
    "offerID": STANDARD_OFFER_ID,
    "offerId": STANDARD_OFFER_ID,
    "userID": STANDARD_USER_ID,
    "userId": STANDARD_USER_ID,
    "propertyID": STANDARD_PROPERTY_ID,
    "propertyId": STANDARD_PROPERTY_ID,
    "listingID": STANDARD_LISTING_ID,
    "listingId": STANDARD_LISTING_ID,
    "placeID": STANDARD_GOOGLE_PLACE_ID,
    "placeId": STANDARD_GOOGLE_PLACE_ID,
    "place_id": STANDARD_GOOGLE_PLACE_ID,
    "googlePlaceID": STANDARD_GOOGLE_PLACE_ID,
    "location_place_id": STANDARD_GOOGLE_PLACE_ID,
}

KeyTypes = Union[int, str]


def _standardize_tags(tags: Optional[Dict[KeyTypes, Any]]) -> Dict[str, Any]:
    standardized_tags = {}
    if not tags:
        return standardized_tags

    for key, value in tags.items():
        string_key = key
        if isinstance(key, int):
            string_key = str(key)

        if string_key in STANDARD_FORMAT_MAPPER:
            standardized_tags[STANDARD_FORMAT_MAPPER[string_key]] = value
        else:
            standardized_tags[string_key] = value

    return standardized_tags


def orjson_default(obj):
    return str(obj)


class ProperLogger:

    def __init__(self, log_location: str, common_tags: Dict[KeyTypes, Any] = None, logger=None):
        """
        Initialize the logger with the basic tags we want logged in every message
        :param log_location: a name for the logical area of code where the log messages are
        being made from. Often this will be the class name like ProperLogger or the filename.
        This isn't auto-generated because the idea is that this is human understandable. The caller
        could pass in __name__ if they so wish.
        :param common_tags: any extra tags that will be added to any subsequent logs
        :param logger: rather than instantiate the logger you can pass one in
        """
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(log_location)

        try:
            level = int(os.environ.get('PROPERLY_LOG_LEVEL', logging.INFO))
        except ValueError:
            level = logging.INFO
        self.logger.level = level

        properly_env = os.environ.get('PROPERLY_STAGE', 'local')
        # Sorting has a performance hit so we only do it outside of prod for test consistency and easier debugging
        self.json_option = orjson.OPT_SORT_KEYS

        self.base_tags = {"log_location": log_location}
        if common_tags:
            self.base_tags = {
                **_standardize_tags(common_tags),
                **self.base_tags
            }

    def _log_with_base_tags(self, level: int, log: Dict[str, Any]):
        self.logger.log(level, orjson.dumps({
            **self.base_tags,
            **log,
        }, option=self.json_option, default=orjson_default))

    def add_common_tags(self, common_tags: Dict[KeyTypes, Any]):
        """
        Add common tags that will continue to be added to subsequent logs
        Replaces any existing tags with the same key
        :param common_tags: extra tags that will be added to any subsequent logs
        """
        self.base_tags = {
            **self.base_tags,
            **_standardize_tags(common_tags),
        }

    def clear_common_tags(self):
        """
        Clears all common tags except log_location
        """
        self.base_tags = {
            "log_location": self.base_tags.get("log_location"),
        }

    def debug(self, message: str, extra_tags: Dict[KeyTypes, Any] = None):
        """
        Log a message along with any tags at a level that will default to being
        ignored in production
        :param message: the message
        :param extra_tags: any extra tags to attach to the log
        """
        extra_tags = _standardize_tags(extra_tags)

        self._log_with_base_tags(logging.DEBUG, {
            **extra_tags,
            "message": message,
        })

    def info(self, message: str, extra_tags: Dict[KeyTypes, Any] = None):
        """
        Log a message along with any tags
        Because of the unstructured nature of this log be considerate about the
        frequency of messages or consider using debug level logging to make it easier
        to silence
        :param message: the message
        :param extra_tags: any extra tags to attach to the log
        """
        extra_tags = _standardize_tags(extra_tags)

        self._log_with_base_tags(logging.INFO, {
            **extra_tags,
            "message": message,
        })

    def metric(self, name: str, value: Any, extra_tags: Dict[KeyTypes, Any] = None):
        """
        Log a metric that can easily be monitored in cloudwatch insights
        :param name: the metric name
        :param value: the metric value
        :param extra_tags: any extra tags to attach to the log
        """
        extra_tags = _standardize_tags(extra_tags)

        self._log_with_base_tags(logging.INFO, {
            **extra_tags,
            "metric_name": name,
            "metric_value": value,
        })

    def warning(self, name: str, message: str, extra_tags: Dict[KeyTypes, Any] = None):
        """
        Log a warning in a structure format that makes it easy to count warnings
        Warnings should be used instead of errors when the severity of the issue is
        smaller but still useful to track
        :param name: the warning name
        :param message: the warning message
        :param extra_tags: any extra tags to attach to the warning log
        """
        extra_tags = _standardize_tags(extra_tags)

        self._log_with_base_tags(logging.WARNING, {
            **extra_tags,
            "warning_name": name,
            "warning_message": message,
        })

    def error(self, name: str, message: str, extra_tags: Dict[KeyTypes, Any] = None):
        """
        Log an error in a structure format that makes it easy to count errors
        :param name: the error name
        :param message: the error message
        :param extra_tags: any extra tags to attach to the error log
        """
        extra_tags = _standardize_tags(extra_tags)

        self._log_with_base_tags(logging.ERROR, {
            **self.base_tags,
            **extra_tags,
            "error_name": name,
            "error_message": message,
        })

    def exception(self, name: str, message: str, extra_tags: Dict[KeyTypes, Any] = None):
        """
        Log an exception in a structured format which will also include a stack trace
        :param name: the exception name
        :param message: the exception message
        :param extra_tags: any extra tags to attach to the exception log
        """
        extra_tags = _standardize_tags(extra_tags)

        self.logger.exception(orjson.dumps({
            **self.base_tags,
            **extra_tags,
            "exception_name": name,
            "exception_message": message,
        }, option=self.json_option, default=orjson_default))



import csv
import decimal
import json
import os
import traceback

from datetime import (
    datetime,
    timezone,
)
from typing import List

import boto3

from properly_util_python.data_utils import ExtendedJsonEncoder


def dynamo_to_dict(dynamo_data):
    if not dynamo_data:
        return dynamo_data
    return json.loads(ExtendedJsonEncoder().encode(dynamo_data))


def dict_to_dynamo(raw_dict):
    dynamo_dict = {}
    for key, val in raw_dict.items():
        if isinstance(val, float):
            dynamo_dict[key] = decimal.Decimal(str(val))
        elif isinstance(val, dict):
            dynamo_dict[key] = dict_to_dynamo(raw_dict[key])
        elif isinstance(val, list):
            new_list = []
            for item in val:
                conversion_dictionary = {'val_to_convert': item}
                result_dictionary = dict_to_dynamo(conversion_dictionary)
                # get + if (new_item) handles dictionary where the items are 'None'
                new_item = result_dictionary.get('val_to_convert')
                if new_item:
                    new_list.append(new_item)
            dynamo_dict[key] = new_list
        elif isinstance(val, datetime):
            if val.tzinfo is None:
                # Timestamps are not required to have timezones.
                # Assume it is in UTC if not specified.
                val = val.replace(tzinfo=timezone.utc)
            milliseconds = int(val.timestamp() * 1000)
            dynamo_dict[key] = milliseconds
        elif val is None:
            # skip null values
            continue
        elif val == "":
            # skip null values
            continue
        else:
            dynamo_dict[key] = val
    return dynamo_dict


def download_files(files=None, bucket_name=None):

    # Without providing explicit keys boto will pick up keys from the environment during dev,
    # and pick up access control from the role in prod
    # local file is ./aws/credentials see: https://docs.aws.amazon.com/cli/latest/userguide/cli-config-files.html
    s3 = boto3.resource('s3', )
    try:
        if not files:
            files = []
        data_dir = 'data/'
        os.makedirs(data_dir, exist_ok=True)

        local_files = []
        for file in files:
            os.makedirs(os.path.dirname(file), exist_ok=True)

        for file in files:
            local_file = file[file.rfind("/") + 1:]
            local_file = data_dir + local_file

            print('downloading:', file)
            print('local_file:', local_file)
            s3.Bucket(bucket_name).download_file(file, local_file)
            local_files.append(local_file)

        return local_files

    except Exception as e:
        print('exception', e)
        print('traceback', traceback.format_exc())


def isfloat(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True


def isint(x):
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b



def load_csv_data(csv_file_name):
    data = []

    with open(csv_file_name, "rt") as f:
        reader = csv.reader(f, delimiter=",")
        for i, line in enumerate(reader):
            data.append(line)
    return data


def csv_to_json(csv_file_name):
    csv_dict = []
    data = load_csv_data(csv_file_name)

    for i, row in enumerate(data[1:]):
        json_row = {}

        for column, value in zip(data[0], row):
            if not column:
                continue
            # id column must be a string
            if any(x in value for x in ['E+','E-','e+','e-']):
                pass
            elif isint(value):
                value = int(value)
            elif isfloat(value):
                value = float(value)
            json_row[column] = value

        # if 'id' not in data[0] or not json_row.get('id',None):
        #     json_row['id'] = i
        # json_row['id'] = str(i)

        csv_dict.append(json_row)

    return csv_dict


def json_to_csv(json_data: List[dict], csv_file_name):

    json_cols = json_data[0]
    json_cols = [key for key in json_cols.keys()]

    json_data_rows = []

    for row in json_data:
        json_data_rows.append([val for val in row.values()])

    with open(csv_file_name, 'w') as csv_file:
        wr = csv.writer(csv_file, quoting=csv.QUOTE_ALL)
        wr.writerow(json_cols)
        wr.writerows(json_data_rows)


if __name__ == '__main__':
    json_data = [
        {'role': 1, 'name': 'Tomiwa'},
        {'role': 2, 'name': 'Craig'},
    ]
    csv_file_name = 'test.csv'
    json_to_csv(json_data, csv_file_name)

    print(csv_to_json(csv_file_name))




import base64
import logging
import json

logger = logging.getLogger()
logger.level = logging.INFO


class KinesisProperlyFormatError(Exception):
    pass


class KinesisHelper:

    def get_data_objects_from_event(self, kinesis_event: dict):
        data_objects_array = []
        records_array = kinesis_event.get('Records', None)
        if not records_array:
            raise RuntimeError('Incorrect event passed to kinesis processing')

        for record in records_array:
            kinesis_block = record.get('kinesis', None)
            if not kinesis_block:
                raise KinesisProperlyFormatError('Invalid Kinesis record')
            data_string = kinesis_block.get('data', None)
            if not data_string:
                raise KinesisProperlyFormatError('Empty Kinesis event')

            data_bytes = base64.standard_b64decode(data_string)

            data_as_decoded_string = data_bytes.decode("utf-8")

            data_as_object = json.loads(data_as_decoded_string)

            data_objects_array.append(data_as_object)

        return data_objects_array


